#ifndef TWN_INPUT_H
#define TWN_INPUT_H

#include "twn_config.h"
#include "twn_vec.h"
#include "twn_util.h"
#include "twn_engine_api.h"

#include <SDL2/SDL.h>

#include <stdint.h>
#include <stdbool.h>


typedef enum ButtonSource {
    BUTTON_SOURCE_NOT_SET,
    BUTTON_SOURCE_KEYBOARD_PHYSICAL,
    BUTTON_SOURCE_KEYBOARD_CHARACTER,
    BUTTON_SOURCE_GAMEPAD,
    BUTTON_SOURCE_MOUSE,
} ButtonSource;


union ButtonCode {
    SDL_Scancode scancode;
    SDL_Keycode keycode;
    SDL_GameControllerButton gamepad_button;
    uint8_t mouse_button;   /* SDL_BUTTON_ enum */
};


/* an input to which an action is bound */
/* it is not limited to literal buttons */
typedef struct Button {
    enum ButtonSource source;
    union ButtonCode code;
} Button;


/* represents the collective state of a group of buttons */
/* that is, changes in the states of any of the bound buttons will affect it */
typedef struct Action {
    size_t num_bindings;

    /* if you bind more than NUM_KEYBIND_SLOTS (set in config.h) */
    /* it forgets the first Button to add the new one at the end */
    Button bindings[NUM_KEYBIND_SLOTS];

    Vec2 position;        /* set if applicable, e.g. mouse click */
    bool is_pressed;
    bool just_changed;
} Action;


typedef struct ActionHashItem {
    char *key;
    Action value;
} ActionHashItem;


typedef struct InputState {
    ActionHashItem *action_hash;
    const uint8_t *keyboard_state; /* array of booleans indexed by scancode */
    uint32_t mouse_state;          /* SDL mouse button bitmask */
    Vec2i mouse_window_position;
    Vec2i mouse_relative_position;
    ButtonSource last_active_source;
    bool is_anything_just_pressed;
} InputState;


TWN_API void input_state_init(InputState *input);
TWN_API void input_state_deinit(InputState *input);
TWN_API void input_state_update(InputState *input);

TWN_API void input_bind_action_scancode(InputState *input,
                                char *action_name,
                                SDL_Scancode scancode);
TWN_API void input_unbind_action_scancode(InputState *input,
                                  char *action_name,
                                  SDL_Scancode scancode);
TWN_API void input_bind_action_mouse(InputState *input,
                             char *action_name,
                             uint8_t mouse_button);
TWN_API void input_unbind_action_mouse(InputState *input,
                               char *action_name,
                               uint8_t mouse_button);

TWN_API void input_add_action(InputState *input, char *action_name);
TWN_API void input_delete_action(InputState *input, char *action_name);

TWN_API bool input_is_action_pressed(InputState *input, char *action_name);
TWN_API bool input_is_action_just_pressed(InputState *input, char *action_name);
TWN_API bool input_is_action_just_released(InputState *input, char *action_name);

TWN_API Vec2 input_get_action_position(InputState *input, char *action_name);

TWN_API void input_set_mouse_captured(InputState *input, bool value);
TWN_API bool input_is_mouse_captured(InputState *input);

#endif
